import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib.colors import LinearSegmentedColormap

PI = np.pi

# FUNCTION: plotpriors
# INPUT: tuple of empirical priors, as created by process.computepriors
# OUTPUT: NULL.  Generates and displays plot in a new figure

def plotpriors(empirical_priors):
    
    displacement_pdf_nm, angle_pdf_nm, displacement_pdf_nc, angle_pdf_nc = empirical_priors
    
    plt.figure(1)
    
    hax = plt.subplot(2,2,1)
    plt.plot(angle_pdf_nc.x, angle_pdf_nc.y, 'k-')
    setcircularplot(hax, False, 1)
    plt.ylabel('Probability density', size=18)
    
    plt.subplot(2,2,2)
    plt.plot(displacement_pdf_nc.x, displacement_pdf_nc.y, 'k-')
    
    hax = plt.subplot(2,2,3)
    plt.plot(angle_pdf_nm.x, angle_pdf_nm.y, 'k-')
    setcircularplot(hax, False, 1)
    plt.xlabel('Angle change (rad)', size=18)
    plt.ylabel('Probability density', size=18)
    
    plt.subplot(2,2,4)
    plt.plot(displacement_pdf_nm.x, displacement_pdf_nm.y, 'k-')
    plt.xlabel('Displacement (length units)', size=18)
    
    # plt.show()
    
# END FUNCTION

# FUNCTION setcircularplot
# INPUTS:
# hax - handle of axes to modify, default gca
# b_symm - bool specifying whether angle distribution is symmetric, default False
# style - int specifying range and interval size
# OUTPUTS: NULL.
    
def setcircularplot(hax=plt.gca(), b_symm=False, style=1):
    plt.axes(hax)
    plot_params = dict(va='center')
    
    if style==1: # max PI, half PI intervals
        if b_symm:
            x_loc = [0, PI/2, PI]
            x_lab = [r"$0$",r"$\pi/2$",r"$\pi$"]
        else:
            x_loc = [-PI, -PI/2, 0, PI/2, PI]
            x_lab = [r"$-\pi$",r"$-\pi/2$",r"$0$",r"$\pi/2$",r"$\pi$"]
    elif style==2: # max PI/2, quarter PI intervals
        if b_symm:
            x_loc = [0, PI/4, PI/2]
            x_lab = [r"$0$",r"$\pi/4$",r"$\pi/2$"]
        else:
            x_loc = [-PI/2, -PI/4,  0, PI/4, PI/2]
            x_lab = [r"$-\pi/2$",r"$-\pi/4$",r"$0$",r"$\pi/4$",r"$\pi/2$"]
    elif style==3: # max PI, third PI interval
        if b_symm:
            x_loc = [0, PI/3, 2*PI/3, PI]
            x_lab = [r"$0$",r"$\pi/3$",r"$2\pi/3$",r"$\pi$"]
        else:
            x_loc = [-PI, -2*PI/3, -PI/3, 0, PI/3, 2*PI/3, PI]
            x_lab = [r"$-\pi$",r"$-2\pi/3$",r"$-\pi/3$",r"$0$",r"$\pi/3$",r"$2\pi/3$",r"$\pi$"]
            
    plt.xticks(x_loc, x_lab, **plot_params)
    plt.xlim([min(x_loc), max(x_loc)])

# END FUNCTION

# FUNCTION redgreencmap
# INPUTS: NULL
# OUTPUTS: colourmap (dict) defining red to green linear map

def redgreencmap():
    cdict = {
        'red':   [(0.0,  1.0, 1.0),
                   (0.5,  1.0, 1.0),
                   (1.0, 0.0, 0.0)],

        'green': [(0.0,  0.0, 0.0),
                  (0.5, 1.0, 1.0),
                   (1.0,  1.0, 1.0)],

        'blue':  [(0.0,  0.0, 0.0),
                   (1.0,  0.0, 0.0)]
             }
    cmap=LinearSegmentedColormap(name='red_to_green', segmentdata=cdict, N=256, gamma=1.0)
    return cmap

# END FUNCTION

# FUNCTION plottrackstateprobs
# INPUTS:
# h_ax - axis handle
# track - [X, Y] track representation, as produced by io.loadtrackscsv
# run_probs - array of run probs for the supplied track, all in [0, 1]
# OUTPUTS: NULL.
# Adds line collection with colours representing state probabilities to the specified axes 

def plottrackstateprobs(h_ax, track, run_probs):
    x = track[0,:]
    y = track[1,:]
    
    # Create a set of line segments so that we can color them individually
    # This creates the points as a N x 1 x 2 array so that we can stack points
    # together easily to get the segments. The segments array for line collection
    # needs to be numlines x points per line x 2 (x and y)
    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    
    # Create the line collection object, setting the colormapping parameters.
    # Have to set the actual values used for colormapping separately.
#    lc = LineCollection(segments, cmap=plt.get_cmap('copper'),
#        norm=plt.Normalize(0, 1))
    lc = LineCollection(segments, cmap=redgreencmap(), norm=plt.Normalize(0, 1))
    lc.set_array(run_probs)
    lc.set_linewidth(2)
    
    h_ax.add_collection(lc)
    
# END FUNCTION

# FUNCTION plotsometrackstateprobs
# Creates a figure and plots a random selection of numtoplot tracks

def plotsometrackstateprobs(tracks_xy, run_probs, numtoplot):
    idx = np.arange(len(tracks_xy))
    np.random.shuffle(idx)
    idx = idx[0:numtoplot-1]
    plt.figure()
    for i in idx:
        if run_probs:
            plottrackstateprobs(plt.gca(), tracks_xy[i], run_probs[i])
        else: # run_probs not supplied
            plottrackstateprobs(plt.gca(), tracks_xy[i], np.ones(np.shape(tracks_xy[i])[1]))
    
    plt.autoscale(enable=True)
    plt.xlabel('x (length units)')
    plt.ylabel('y (length units)')
    # plt.show()
    
# END FUNCTION

# FUNCTION
def plotsimulationtransitionprobs(noiselevels, full_bs_tps, so_bs_tps, true01, true10):

    noiselevels = np.array(noiselevels)
    N = len(noiselevels)

    full_bs01 = [sorted(0.02/np.array(full_bs_tps[i])[:,0]) for i in range(N)]
    full_bs10 = [sorted(0.02/np.array(full_bs_tps[i])[:,1]) for i in range(N)]
    so_bs01 = [sorted(0.02/np.array(so_bs_tps[i])[:,0]) for i in range(N)]
    so_bs10 = [sorted(0.02/np.array(so_bs_tps[i])[:,1]) for i in range(N)]

    full_mean01 = [np.mean(x) for x in full_bs01]
    full_mean10 = [np.mean(x) for x in full_bs10]
    so_mean01 = [np.mean(x) for x in so_bs01]
    so_mean10 = [np.mean(x) for x in so_bs10]
    
    full_ci01 = np.transpose( [ [x[24],x[974]] for x in full_bs01 ] )
    full_ci10 = np.transpose( [ [x[24],x[974]] for x in full_bs10 ] )
    so_ci01 = np.transpose( [ [x[24],x[974]] for x in so_bs01 ] )
    so_ci10 = np.transpose( [ [x[24],x[974]] for x in so_bs10 ] )
    
    for i in range(N):
        full_ci01[0,i] = full_mean01[i]-full_ci01[0,i]
        full_ci01[1,i] -= full_mean01[i]
        full_ci10[0,i] = full_mean10[i]-full_ci10[0,i]
        full_ci10[1,i] -= full_mean10[i]
        so_ci10[0,i] = so_mean10[i]-so_ci10[0,i]
        so_ci10[1,i] -= so_mean10[i]
        so_ci01[0,i] = so_mean01[i]-so_ci01[0,i]
        so_ci01[1,i] -= so_mean01[i]
    
    f, axarr = plt.subplots(2, sharex=True)
    ax1 = axarr[0]
    ax3 = axarr[1]
    
    ax2 = ax1.twinx()
    ax1.errorbar(noiselevels, full_mean01, full_ci01, fmt='k-', linewidth=2, capsize=5)
    ax2.errorbar(noiselevels+0.01, full_mean10, full_ci10, fmt='r-', linewidth=2, capsize=5, alpha=0.5)
    
    ax4 = ax3.twinx()
    ax3.errorbar(noiselevels, so_mean01, so_ci01, fmt='k-', linewidth=2, capsize=5)
    ax4.errorbar(noiselevels+0.01, so_mean10, so_ci10, fmt='r-', linewidth=2, capsize=5)
    
    ax1.plot(noiselevels, np.ones(np.shape(noiselevels))*true01, 'k--')
    ax2.plot(noiselevels, np.ones(np.shape(noiselevels))*true10, 'r--')
    ax3.plot(noiselevels, np.ones(np.shape(noiselevels))*true01, 'k--')
    ax4.plot(noiselevels, np.ones(np.shape(noiselevels))*true10, 'r--')

    for tl in ax2.get_yticklabels():
        tl.set_color('r')
    for tl in ax4.get_yticklabels():
        tl.set_color('r')

    # 0.1, 1
    ytick01 = np.arange(0.05, 0.4, 0.05)
    ytick10 = np.arange(0.5, 3.5, 0.5)
    ytick01lim = [min(ytick01), max(ytick01)]
    ytick10lim = [min(ytick10), max(ytick10)]
    ax1.set_ylim(ytick01lim)
    ax1.set_yticks(ytick01)
    ax2.set_ylim(ytick10lim)
    ax2.set_yticks(ytick10)
    ax3.set_ylim(ytick01lim)
    ax3.set_yticks(ytick01)
    ax4.set_ylim(ytick10lim)
    ax4.set_yticks(ytick10)

    # 0.25, 1
#    ytick01 = np.arange(0.2, 0.31, 0.1)
#    ytick10 = np.arange(0.9, 1.3, 0.1)
#    
#    ax1.set_ylim([0.2, 0.3])
#    ax1.set_yticks(ytick01)
#    ax2.set_ylim([0.7, 1.2])
#    ax2.set_yticks(ytick10)
#    ax3.set_ylim([0.2, 0.3])
#    ax3.set_yticks(ytick01)
#    ax4.set_ylim([0.7, 1.2])
#    ax4.set_yticks(ytick10)    

    ax1.set_xlim([min(noiselevels)-0.01, max(noiselevels)*1.02])
    ax3.set_xlim([min(noiselevels)-0.01, max(noiselevels)*1.02])
    
    plt.subplots_adjust(left=0.13,right=0.88)
    ax3.set_xlabel(r'Noise level ($\mu\mathrm{m}^2\mathrm{s}^{-1}$)')
    ax1.set_ylabel(r'$\tau_{12}$')
    ax3.set_ylabel(r'$\tau_{12}$')
    ax2.set_ylabel(r'$\tau_{21}$', color='r')
    ax4.set_ylabel(r'$\tau_{21}$', color='r')
    # plt.show()

# END FUNCTION

# FUNCTION
def fpandfnrate(noiselevels, fprs, fnrs, fpmax, fnmax):
    markerorder = ['kx','k+','kD','ks','k+','k^']
    markersize=10
    n = len(fprs)
    fig=plt.figure()
    
    f, axarr = plt.subplots(2, sharex=True)
    ax1 = axarr[0]
    ax2 = axarr[1]
    
    
    [ax1.plot(noiselevels, fprs[i], 'k--') for i in range(n)]
    [ax2.plot(noiselevels, fnrs[i], 'k--') for i in range(n)]
    [ax1.plot(noiselevels, fprs[i], markerorder[i], markersize=markersize, markerfacecolor='none', mec='k') for i in range(n)]
    [ax2.plot(noiselevels, fnrs[i], markerorder[i], markersize=markersize, markerfacecolor='none', mec='k') for i in range(n)]
    
    ax1.set_xlim([0, max(noiselevels)*1.02])
    ax2.set_xlim([0, max(noiselevels)*1.02]) 
    ax1.set_ylim([0, fpmax])
    ax2.set_ylim([0, fnmax])

    ax2.set_xlabel(r"Noise level ($\mu\mathrm{m}^2\mathrm{s}^{-1}$)", fontsize=18)
    ax1.set_ylabel(r"False positive rate", fontsize=18)
    ax2.set_ylabel(r"False negative rate", fontsize=18)
    # plt.show()

# END FUNCTION

# FUNCTION
def plotmino(nems, mac):

    # Set the min/max of the data
    xmin = 0
    xmax = 1
    ymin = 0
    ymax = np.pi

    # Make the 'main' temperature plot
    xbins = np.linspace(start = xmin, stop = xmax, num = 20)
    ybins = np.linspace(-np.pi, np.pi, num = 20)
    xcenter = (xbins[0:-1]+xbins[1:])/2.0
    ycenter = (ybins[0:-1]+ybins[1:])/2.0
    aspectratio = 1.0*(xmax - 0)/(1.0*ymax - 0)
    H, xedges,yedges = np.histogram2d(nems, mac, bins=(ybins, xbins))
    X = xcenter
    Y = ycenter
    Z = H

    plt.figure()
    plt.pcolor(X, Y, Z, cmap='gray')
    plt.xlabel('NEMS', fontsize=18)
    plt.ylabel('MAC', fontsize=18)
    plt.colorbar()
    # plt.show()


# FUNCTION twodimhistobservationpdf
# creates a 2D histogram of the observation pdf with marginal distributions at each edge
def twodimhistobservationpdf(tracks_rt, nxbins=50, nybins=50):
    
    from matplotlib.ticker import NullFormatter
    from matplotlib.ticker import MaxNLocator
    
    r = np.concatenate( [x[0,1:] for x in tracks_rt] )
    t = np.concatenate( [x[1,1:] for x in tracks_rt] )
    
    bw = True
    contours = False

    # Set up your x and y labels
    xlabel = '$\mathrm{Framewise\\ displacement\\ }(\mu m)$'
    ylabel = '$\mathrm{Framewise\\ angle\\ change\\ (rad)}$'

    # Define the locations for the axes
    left, width = 0.12, 0.55
    bottom, height = 0.12, 0.55
    bottom_h = left_h = left+width+0.02

    # Set up the geometry of the three plots
    rect_temperature = [left, bottom, width, height] # dimensions of temp plot
    rect_histx = [left, bottom_h, width, 0.25] # dimensions of x-histogram
    rect_histy = [left_h, bottom, 0.25, height] # dimensions of y-histogram

    # Set up the size of the figure
    fig = plt.figure(1, figsize=(9.5,9))

    # Make the three plots
    axTemperature = plt.axes(rect_temperature) # temperature plot
    axHistx = plt.axes(rect_histx) # x histogram
    axHisty = plt.axes(rect_histy) # y histogram

    # Remove the inner axes numbers of the histograms
    nullfmt = NullFormatter()
    axHistx.xaxis.set_major_formatter(nullfmt)
    axHisty.yaxis.set_major_formatter(nullfmt)

    # Find the min/max of the data
    xmin = 0
    xmax = max(r)
    ymin = -np.pi
    ymax = np.pi

    # Make the 'main' temperature plot
    xbins = np.linspace(start = xmin, stop = xmax, num = nxbins)
    ybins = np.linspace(-np.pi, np.pi, num = nybins)
    xcenter = (xbins[0:-1]+xbins[1:])/2.0
    ycenter = (ybins[0:-1]+ybins[1:])/2.0
    aspectratio = 1.0*(xmax - 0)/(1.0*ymax - 0)
    H, xedges,yedges = np.histogram2d(t,r,bins=(ybins,xbins))
    X = xcenter
    Y = ycenter
    Z = H

    # Plot the temperature data
#    cax = axTemperature.imshow(H, extent=[xmin,xmax,ymin,ymax], \
#                               interpolation='nearest', origin='lower',aspect=aspectratio, cmap=plt.cm.gist_yarg)
    cax = axTemperature.pcolor(X,Y,H,cmap=plt.cm.gist_yarg,vmin=0,vmax=100)
    cbar = fig.colorbar(cax)

    #Plot the axes labels
    axTemperature.set_xlabel(xlabel,fontsize=25)
    axTemperature.set_ylabel(ylabel,fontsize=25)

    #Make the tickmarks pretty
    ticklabels = axTemperature.get_xticklabels()
    for label in ticklabels:
        label.set_fontsize(18)
        label.set_family('serif')

    ticklabels = axTemperature.get_yticklabels()
    for label in ticklabels:
        label.set_fontsize(18)
        label.set_family('serif')

    #Set up the plot limits
    axTemperature.set_xlim([xmin, xmax])
    axTemperature.set_ylim([ymin, ymax])

    #Plot the histograms
    axHistx.hist(r, bins=xbins, color = 'silver')
    axHisty.hist(t, bins=ybins, orientation='horizontal', color = 'dimgray')

    #Set up the histogram limits
    axHistx.set_xlim( 0, max(r) )
    axHisty.set_ylim( min(t), max(t))

    #Make the tickmarks pretty
    ticklabels = axHistx.get_yticklabels()
    for label in ticklabels:
        label.set_fontsize(12)
        label.set_family('serif')

    #Make the tickmarks pretty
    ticklabels = axHisty.get_xticklabels()
    for label in ticklabels:
        label.set_fontsize(12)
        label.set_family('serif')

    #Cool trick that changes the number of tickmarks for the histogram axes
    axHisty.xaxis.set_major_locator(MaxNLocator(4))
    axHistx.yaxis.set_major_locator(MaxNLocator(4))

    #if(filename):
    #savefig(filename + '.eps',format = 'eps', transparent=True)
    #savefig(filename + '.pdf',format = 'pdf', transparent=True)
    #savefig(filename + '.png',format = 'png', transparent=True)
    # plt.show()
