import numpy as np
from scipy.stats.kde import gaussian_kde
from scipy.interpolate import interp1d
import scipy.ndimage.measurements as meas
from scipy import optimize
from itertools import izip, imap
# import Miniball

PI = np.pi

# define the number of gridpoints for r and theta discretisation
num_t_grid_pts = 256
num_r_grid_pts = 256

# FUNCTION: trackcarttopolarabsolute
# INPUT: list of tracks, each is a 2xN matrix with X and Y coords
# OUTPUT: list of tracks in displacement, ABSOLUTE angle representation

def trackcarttopolarabsolute(tracks_in):
    n = len(tracks_in)
    quantities_out = [0]*n
    for i in range(n):
        track = tracks_in[i]
        dx = np.diff(track[0,:])
        dy = np.diff(track[1,:])
        displacement = np.sqrt(dx*dx+dy*dy)
        angle = np.arctan2(dy, dx)
        quantities_out[i] = np.vstack( (displacement, angle) )
    return quantities_out

# END FUNCTION

# FUNCTION: trackcarttopolar
# INPUT: list of tracks, each is a 2xN matrix with X and Y coords
# OUTPUT: list of tracks in displacement, RELATIVE angle representation

def trackcarttopolar(tracks_in):
    n = len(tracks_in)
    quantities_out = [0 for i in range(n)]
    for i in range(n):
        track = tracks_in[i]
        dx = np.diff(track[0,:])
        dy = np.diff(track[1,:])
        displacement = np.sqrt(dx*dx+dy*dy)
        angle = np.arctan2(dy, dx)
        angle_diff = np.diff(angle)
        angle_diff[angle_diff>(PI)] = angle_diff[angle_diff>(PI)]-2*PI
        angle_diff[angle_diff<(-PI)] = angle_diff[angle_diff<(-PI)]+2*PI
        # insert zero at start of angles
        angle_diff = np.append(0., angle_diff)
        quantities_out[i] = np.vstack( (displacement, angle_diff) )
    return quantities_out

# END FUNCTION

# FUNCTION: computepriors
# INPUT: tracks from the non-motile and non-chemotactic mutants, in the form output by io.loadtrackscsv
# OUTPUT: tuple of 4 linear interpolation objects corresponding to displacement pdf NM, displacement pdf NC, angle pdf NM, angle pdf NC

def computepriors(tracks_nm, tracks_nc, *tracks_wt):
    
    eps = 1e-12
    
    # check inputs
    if len(tracks_wt)==1:
        btrackssupplied = True
        quantities_wt = trackcarttopolar(tracks_wt[0])
    else:
        btrackssupplied = False
    
    # compute displacements and angles
    quantities = trackcarttopolar(tracks_nc)
    
    displacements_nc = angles_nc = np.array([])
    
    for q in quantities:
        d = q[0,:]
        a = q[1,1:]
        displacements_nc = np.concatenate([displacements_nc, d])
        angles_nc = np.concatenate([angles_nc, a])
    
    quantities = trackcarttopolar(tracks_nm)
    
    displacements_nm = angles_nm = np.array([])
    
    for q in quantities:
        d = q[0,:]
        a = q[1,1:]    
        displacements_nm = np.concatenate([displacements_nm, d])
        angles_nm = np.concatenate([angles_nm, a])
       
    # obtain KDE for each dataset
        
    displacement_pdf_nc = gaussian_kde(displacements_nc)
    angle_pdf_nc = gaussian_kde(angles_nc)
    displacement_pdf_nm = gaussian_kde(displacements_nm)
    angle_pdf_nm = gaussian_kde(angles_nm)
    
    # find maximum displacement
    rho_nm = np.max(displacements_nm)
    rho_nc = np.max(displacements_nc)
    rho = max([rho_nc, rho_nm])
    
    # evaluating KDE repeatedly is slow, so form a linear interpolation instead
    r = np.linspace(0,rho,num_r_grid_pts)
    theta = np.linspace(-PI,PI,num_t_grid_pts)
    
    # compute values for lin interp object, then impose a numerical minimum
    a_nc = angle_pdf_nc(theta)
    a_nm = angle_pdf_nm(theta)
    d_nm = displacement_pdf_nm(r)
    a_nc[a_nc<eps]=eps
    a_nm[a_nm<eps]=eps
    d_nm[d_nm<eps]=eps

    angle_pdf_nc = interp1d(theta, a_nc, kind='linear', copy=False)
    angle_pdf_nm = interp1d(theta, a_nm, kind='linear', copy=False)
    displacement_pdf_nm = interp1d(r, d_nm, kind='linear', copy=False, bounds_error=False, fill_value=0.0)
    
    # if additional tracks supplied, compute rho from these too
    if btrackssupplied:
        
        rho_wt = 0.
        displacements_wt = np.array([])
        
        for q in quantities_wt:
            d = q[0,:]
            displacements_wt = np.concatenate([displacements_wt, d])
        
        rho_wt = max(displacements_wt)
        print "rho = ", rho, "rho_wt = ", rho_wt
            
        if rho_wt > rho:
            # compute difference between the maxima
            rho_diff = rho_wt-rho
            # compute the proportion of WT tracks above rho
            proportion_outliers = float(sum(displacements_wt>rho)) / float(np.size(displacements_wt))
            print "proportion outliers = ", proportion_outliers
            # add extra density to pdf to make it NORMALIZED
            x = np.concatenate([ r, [rho_wt] ])
            d_nc = np.concatenate([ (1-proportion_outliers)*displacement_pdf_nc(r[0:-1]), proportion_outliers/rho_diff*np.array([1,1]) ])
            # impose minimum
            d_nc[d_nc<eps] = eps
            displacement_pdf_nc = interp1d(x, d_nc, kind='linear', copy=False )
        else:
            # fill value is 1.0 NOT NORMALIZED!!
            d_nc = displacement_pdf_nc(r)
            # impose minimum
            d_nc[d_nc<eps] = eps
            displacement_pdf_nc = interp1d(r, d_nc, kind='linear', copy=False, bounds_error=False, fill_value=1.0)
            
    else:
        # fill value is 1.0 NOT NORMALIZED!!
        d_nc = displacement_pdf_nc(r)
        # impose minimum
        d_nc[d_nc<eps] = eps
        displacement_pdf_nc = interp1d(r, d_nc, kind='linear', copy=False, bounds_error=False, fill_value=1.0)
    
    empirical_priors = displacement_pdf_nm, angle_pdf_nm, displacement_pdf_nc, angle_pdf_nc
    
    return empirical_priors

# END FUNCTION

# FUNCTION thresholdrunprobabilities
# INPUTS:
# run_probs - list of state probabilities, generated by hmm.Hmm or similar
# minimumrunduration - int >=0 describing the min permissible run duration
# minimumstopduration - int >=0 describing the min permissible stop duration
# OUTPUTS:
# stop_status - list of run STATUS arrays, containing 1 for a run and 0 for a stop 

def thresholdrunprobabilities(run_probs, minimumrunduration=1, minimumstopduration=1):
    
    assert (minimumstopduration>0) and ( type(minimumstopduration) is type(int())), 'minimumstopduration must be an integer >=0'
    assert (minimumrunduration>0) and ( type(minimumrunduration) is type(int())), 'minimumrunduration must be an integer >=0'

    if (minimumrunduration<=1) and (minimumstopduration<=1): # trivial case
        
        return [np.where(probs>=0.5, 1, 0) for probs in run_probs]
    
    else: # need to apply thresholding
        
        stop_status = []
        
        for probs in run_probs: # loop over tracks
    
            # if minimum run durations have been supplied, enforce these
            if minimumrunduration>1:
                
                # segment runs
                labels, numruns = meas.label(probs>=0.5)
                            
                # flip where necessary
                rundurations = np.array([sum(labels==x) for x in range(1,numruns+1)])
                for i in np.where(rundurations<minimumrunduration)[0]+1:
                    labels[labels==i] = 0
#                    print 'flipping run idx', i, 'with duration', rundurations[i-1]
                    
                # segment stops
                labels, numstops = meas.label(labels==0)
                
            else: # else just skip straight to stop segmentation
                
                labels, numstops = meas.label(probs<0.5)
                
            
            
            if minimumstopduration>1:
                
                # flip where necessary
                stopdurations = np.array([sum(labels==x) for x in range(1,numstops+1)])
                for i in np.where(stopdurations<minimumstopduration)[0]+1:
                    labels[labels==i] = 0 
#                    print 'flipping stop idx', i, 'with duration', stopdurations[i-1] 
                    
            stop_status.append(np.where(labels, 0, 1))
            
        return stop_status

# END FUNCTION        

# FUNCTION computesegmenteddispandangle
# INPUTS:
# tracks_xy - list of tracks as generated by io.loadtrackscsv
# run_probs - list of state probabilities, generated by hmm.Hmm
# OUTPUTS:
# Tuple of 4 arrays containing run disp, stop disp, run angles, stop angles
# Separates the displacements and angle changes in states labelled run and stop.

def computesegmenteddispandangle(tracks_xy, run_probs):
    
    checktracksandrunprobs(tracks_xy, run_probs)
    tracks_rt = trackcarttopolar(tracks_xy)
    run_displacements = []
    stop_displacements = []
    run_angles = []
    stop_angles = []
    
    for track, probs in izip(tracks_rt, run_probs):
        assert np.size(track,1)==np.size(probs), "Track and state probabilities are not the same size."

        stop_indices = np.where(probs<0.5)[0]
        run_indices = np.where(probs>=0.5)[0]
        
        # extract disps from both regions
        this_stop_disp = track[0,stop_indices]
        this_run_disp = track[0,run_indices]
        
        # extract angles from both regions
        # NB first angle is always zero so discard any reference to this
        stop_indices = stop_indices[stop_indices!=0]
        run_indices = run_indices[run_indices!=0]
        
        this_stop_angle =  track[1,stop_indices]
        this_run_angle =  track[1,run_indices]
        
        run_displacements = np.concatenate((run_displacements, this_run_disp))
        stop_displacements = np.concatenate((stop_displacements, this_stop_disp))
        run_angles = np.concatenate((run_angles, this_run_angle))
        stop_angles = np.concatenate((stop_angles, this_stop_angle))
        
    return run_displacements, stop_displacements, run_angles, stop_angles

# END FUNCTION

# FUNCTION computesegmentdurations
# INPUTS:
# tracks_xy - list of tracks as generated by io.loadtrackscsv
# run_probs - list of state probabilities, generated by hmm.Hmm
# OUTPUTS:
# Tuple of 2 arrays containing durations of inferred run and stop events

def computesegmentdurations(tracks_xy, run_probs):
    
    checktracksandrunprobs(tracks_xy, run_probs)
    tracks_rt = trackcarttopolar(tracks_xy)
    run_durations = []
    stop_durations = []
        
    for track, probs in izip(tracks_rt, run_probs):
        assert np.size(track,1)==np.size(probs), "Track and state probabilities are not the same size."
        # segment run_probs
        labels, num_stops = meas.label(np.where(probs<0.5, 1, 0))
#        stop_indices = [where(labels==i) for i in np.arange(max(labels))]
        this_stop_dur = [sum(labels==i) for i in np.arange(max(labels))]
        labels, num_runs = meas.label(np.where(probs>=0.5, 1, 0))
        this_run_dur = [sum(labels==i) for i in np.arange(max(labels))]
        
        stop_durations = np.append(stop_durations, this_stop_dur)
        run_durations = np.append(run_durations, this_run_dur)
        
    return run_durations, stop_durations

# END FUNCTION

# FUNCTION computestopwisequantities
# INPUTS:
# tracks_xy - list of tracks as generated by io.loadtrackscsv
# run_probs - list of state probabilities, generated by hmm.Hmm
# [angleframeave] - optional value specifying the number of transitions over which to average to obtain stopwise entry and exit angles/speeds.  This is the MAX value used; where fewer transitions are available, as many as possible are used.
# [minimumstopduration] - optional value defining the minimum permissible stop duration
# [minimumrunduration] - optional value defining the minimum permissible run duration
# OUTPUTS:
# Tuple of 5 arrays:
# stopidx - sequence of arrays containing the indices of the identified stopping phase.  The length of each sequence element gives the number of inferred stops in that track.
# entryfwdisplacement - sequence of sequences, each of which is a list of the observed entry displacement before an identified stop.
# exitfwdisplacement - sequence of sequences, each of which is a list of the observed exit displacement following an identified stop.
# swanglechange - sequence of sequences, each of which is a list of the observed angle change during an identified stop.
# swduration - sequence of sequences, each of which is a list of the observed duration of an identified stop in numbers of frames.

def computestopwisequantities(tracks_xy, run_probs, **kwargs):
    
    checktracksandrunprobs(tracks_xy, run_probs)
    
    # unpack params from kwargs
    angleframeave = kwargs.pop('angleframeave', 1)
    minimumstopduration = kwargs.pop('minimumstopduration', 1)
    minimumrunduration = kwargs.pop('minimumrunduration', 1)
    
    assert (angleframeave>=1) and ( type(angleframeave) is type(int()) ), 'angleframeave must be an integer >=1'
    
    # impose duration thresholds
    stop_status = thresholdrunprobabilities(run_probs, minimumrunduration, minimumstopduration)
    
    tracks_absolute_rt = trackcarttopolarabsolute(tracks_xy)
    
    #  initialise output lists
    stopidx = []
    entryfwdisplacement = []
    exitfwdisplacement = []
    swanglechange = []
    swduration = []
    
    for track, probs in izip(tracks_absolute_rt, stop_status):
        
        labels, numstops = meas.label(probs)
                
        # discard stops that are not pre or proceeded by a run
        if labels[0]==1:
            labels[labels==1] = 0
            labels[labels!=0] = labels[labels!=0]-1
            numstops -= 1
#            print 'discarded first stop'
        
        if labels[-1]==numstops:
            labels[labels==numstops] = 0
            numstops -= 1
#            print 'discarded last stop'
            
        this_stopidx = [0]*numstops
        this_entryfwdisplacement = list(this_stopidx)
        this_exitfwdisplacement = list(this_stopidx)
        this_swanglechange = list(this_stopidx)
        this_swduration = list(this_stopidx)
      
        # loop over each stop phase
        for i in range(1,numstops+1):
            
            # find initial and final stop index
            temp_stopidx = np.where(labels==i)[0]
            initialidx = temp_stopidx[0]
            finalidx = temp_stopidx[-1]
            
            # compute entry and exit indices
            if i>1: # previous stop exists
                prev_stopidx = np.where(labels==i-1)[0][-1]
            else:
                prev_stopidx = -1
                
            if i<numstops: # later stop exists
                next_stopidx = np.where(labels==i+1)[0][0]
            else:
                next_stopidx = len(probs)+1
                
            entryidx = range( max(0, initialidx-angleframeave, prev_stopidx+1), initialidx )
#            print 'finalidx:', finalidx
#            print 'next_stopidx:', next_stopidx
            exitidx = range( finalidx, min( len(probs), finalidx+angleframeave, next_stopidx-1 ) )
            
            # compute quantities of interest
            this_entryfwdisplacement[i-1] = np.mean(track[0, entryidx])
            this_exitfwdisplacement[i-1] = np.mean(track[0, exitidx])
            this_swanglechange[i-1] = np.mean(track[1, exitidx]) - np.mean(track[1, entryidx])
            if this_swanglechange[i-1]>PI:
                this_swanglechange[i-1] = this_swanglechange[i-1]-2*PI
            if this_swanglechange[i-1]<(-PI):
                this_swanglechange[i-1] = this_swanglechange[i-1]+2*PI
                    
#            print '1:', this_swanglechange[i-1]
            this_swduration[i-1] = len(temp_stopidx)
#            print 'this_swduration[i-1]=', this_swduration[i-1]
            this_stopidx[i-1] = temp_stopidx
        
#        print '2:', this_swanglechange 
        
        stopidx.append(this_stopidx)
        entryfwdisplacement.append(this_entryfwdisplacement)
        exitfwdisplacement.append(this_exitfwdisplacement)
        swanglechange.append(this_swanglechange)
        swduration.append(this_swduration)
        
        # end loop over tracks 
        
    return stopidx, entryfwdisplacement, exitfwdisplacement, swanglechange, swduration

# END FUNCTION

# FUNCTION discretiseunderlyingstatesequence
# Discretises a true continuous time state sequence for comparison with inferred sequence.
# INPUTS:
# eventtimes - sequence of sequences containing the times at which events occur, as generated by the simulation method.  Events are alternately runs and stops.
# dt - sampling interval
# numtimepoints - the number of sampling intervals
# OUTPUTS:
# runprobs - sequence of sequences, equivalent to that outputted by the HMM

def discretiseunderlyingstatesequence(eventtimes, sampletimes):
    runprobs = []
    # loop over each supplied track
    for i in range(len(eventtimes)):
        
        thiseventtimes = eventtimes[i]
        numevents = len(thiseventtimes)
        numsampletimes = np.size(sampletimes)
        thisrunprobs = np.ones(numsampletimes-1)
        
        # loop over event pairs, assuming the first event is a stop
        for j in range(int(numevents/2)):
            
            starttime = thiseventtimes[2*j]
            endtime = thiseventtimes[2*j+1]
            stopidx = np.where([all(t) for t in zip(sampletimes>starttime, sampletimes<=endtime)])[0]
            
            if len(stopidx)==0: # event falls BETWEEN sample points
#                print 'Event between sample points'
                stopidx = np.argmax(sampletimes>starttime)-1
            else: # event covers at least one whole interval
#                print 'Event covers multiple sample points'
                stopidx = np.insert(stopidx, 0, stopidx[0]-1)
            
#            print starttime, endtime, stopidx
            thisrunprobs[stopidx] = 0.
            
        # check for ending stop events too
        if np.remainder(numevents, 2) != 0:
            starttime = thiseventtimes[-1]
#            print 'Final stopping event at t=', starttime
            stopidx = np.argmax(sampletimes>starttime)-1
#            print 'stopidx', stopidx, 'until end of track'
            thisrunprobs[stopidx:] = 0.
#            print
            
#        print '*******'
        runprobs.append(thisrunprobs)
    
    return runprobs

# END FUNCTION

# FUNCTION compareinferredandtruesequences
# Compares the inferred state sequence with the underlying sequence
# INPUTS:
# obsrunprobs - sequence containing inferred run probabilities
# truerunstatus - sequence containing true state (all entries in {1, 0})
# OUTPUTS:
# numfalsepositives, numfalsenegatives
# matchingstopindex - Nx2 array, each row corresponding to a matched stopping event, left col is the OBS sequence, right col is the TRUE sequence.  Each entry gives the index of the stop in that state sequence.

def compareinferredandtruesequences(obsrunprobsin, truerunstatusin):
    
    # create working copies
    obsrunprobs = np.copy(obsrunprobsin)
    truerunstatus = np.copy(truerunstatusin)
    
    # check inputs
    assert np.any([len(t[0])==len(t[1]) for t in zip(obsrunprobs, truerunstatus)]), "Length of supplied sequences is incompatible"
    
    # init outputs
    numfp = []
    numfn = []
    matchingstopidx = []
    numtrue = []
    
    # convert obs run probs to status
    obsstopstatus = thresholdrunprobabilities(obsrunprobs, 1, 1)
    
    # loop over pairs of track sequences
    for truestat, obsstat in izip(truerunstatus, obsstopstatus):

        # segment true underlying state
        truelabels, truenumstops = meas.label(np.where(truestat,0,1))
        truestopslice = meas.find_objects(truelabels)
        truestopslice = [x[0] for x in truestopslice]
        # segment obs underlying state
        obslabels, obsnumstops = meas.label(np.where(obsstat,0,1))
        obsstopslice = meas.find_objects(obslabels)
        obsstopslice = [x[0] for x in obsstopslice]

    	# if no observed stops, finish here
    	if obsnumstops == 0:
    	    matchingstopidx.append([])
    	    numfp.append(0)
    	    numfn.append(truenumstops)
            numtrue.append(truenumstops)
    	    continue
        
        # if no true stops, finish here
        if truenumstops == 0:
            matchingstopidx.append([])
            numfp.append(obsnumstops)
            numfn.append(0)
            numtrue.append(0)
            continue
                
        # overlap matrix
        overlaps = np.zeros((obsnumstops, truenumstops), dtype=int)
        
        for i in range(obsnumstops):
            
            obsstop0 = int( obsstopslice[i].start )
            obsstop1 = int( obsstopslice[i].stop )
            
            for j in range(truenumstops):
                
                truestop0 = int( truestopslice[j].start )
                truestop1 = int( truestopslice[j].stop )
                
                # test whether any of the obsstops overlap
                overlapcond=[]
                overlapcond.append(obsstop1>truestop0 and obsstop1<=truestop1) # obs endpoint inside true stop
                overlapcond.append(obsstop0>=truestop0 and obsstop0<truestop1) # obs startpoint inside true stop
                overlapcond.append(obsstop0<=truestop0 and obsstop1>=truestop1) # obs stop fully overlaps true stop
                
                if np.any(overlapcond):
                    
                    overlaps[i,j] = 1
                    
#        ipdb.set_trace()
        # use overlaps matrix to extract results
        
        obssum = np.sum(overlaps, axis=1)
        truesum = np.sum(overlaps, axis=0)
        
        numfp.append(np.sum(obssum==0))
        numfn.append(np.sum(truesum==0))
        numtrue.append(truenumstops)
        
        # check whether either obs or true first/last stop is at the very end of the track and remove if so
        bregenobs = False
        bregentrue = False
        bremovedobsstart = False
        bremovedtruestart = False
        
        if obsstopslice[0].start == 0:
            obsstat[ obsstopslice[0] ] = 0
            bregenobs = True
            bremovedobsstart = True
        
        if obsstopslice[-1].stop == len(obsstat):
            obsstat[ obsstopslice[-1] ] = 0
            bregenobs = True
            
        if truestopslice[0].start == 0:
            truestat[ truestopslice[0] ] = 0
            bregentrue = True
            bremovedtruestart = True
            
        if truestopslice[-1].stop == 0:
            truestat[ truestopslice[-1] ] = 0
            bregentrue = True
                        
        if bregenobs: # regenerate obs labels
            obslabels, obsnumstops = meas.label(np.where(obsstat,0,1))
            obsstopslice = meas.find_objects(obslabels)
            obsstopslice = [x[0] for x in obsstopslice]
                
        if bregentrue:
            truelabels, truenumstops = meas.label(np.where(truestat,0,1))
            truestopslice = meas.find_objects(truelabels)
            truestopslice = [x[0] for x in truestopslice]
                 
        # regenerate overlaps matrix
    
        overlaps = np.zeros((obsnumstops, truenumstops), dtype=int)
        
        for i in range(obsnumstops):
            
            obsstop0 = int( obsstopslice[i].start )
            obsstop1 = int( obsstopslice[i].stop )
            
            for j in range(truenumstops):
                
                truestop0 = int( truestopslice[j].start )
                truestop1 = int( truestopslice[j].stop )
                
                # test whether any of the obsstops overlap
                overlapcond=[]
                overlapcond.append(obsstop1>truestop0 and obsstop1<=truestop1) # obs endpoint inside true stop
                overlapcond.append(obsstop0>=truestop0 and obsstop0<truestop1) # obs startpoint inside true stop
                overlapcond.append(obsstop0<=truestop0 and obsstop1>=truestop1) # obs stop fully overlaps true stop
                
                if np.any(overlapcond):
                    
                    overlaps[i,j] = 1
    
        obssum = np.sum(overlaps, axis=1)
        truesum = np.sum(overlaps, axis=0)
    
        # reduce overlaps matrix to include only UNIQUE and genuine matches
        obsuniqueidx = np.where(obssum==1)[0]
        
        # store unique matches
        uniquematches = []
        
        for i in obsuniqueidx:
            # check whether this match is indeed unique
            temp = np.where(overlaps[i,:]==1)[0]
            if truesum[temp]==1: # unique match
                uniquematches.append([i, temp])
        
        uniquematches = np.array(uniquematches)
        
        matchingstopidx.append(uniquematches)
        
        # fix indices if initial stop is invalid
    #    if bremovedobsstart:
    #        uniquematches[:,0] += 1
    #    
    #    if bremovedtruestart:
    #        uniquematches[:,1] += 1

    obsrunprobs[0]=1e6
    truerunstatus[0]=1e6
    return numfp, numfn, numtrue, matchingstopidx

# END FUNCTION

# FUNCTION computeminbounddiameter
# Computes the smallest circular diameter encompassing all points in each supplied track
# INPUTS:
# tracks
# OUTPUTS:
# mbd - list with MBD for each track

# DISABLED as it requires the C++ module Miniball

# def computeminbounddiameter(tracks):
#
#     mbd = []
#
#     for track in tracks:
#         # convert tracks list to vector format
#         x = Miniball.VectorOfDouble()
#         y = Miniball.VectorOfDouble()
#         x.extend(track[0,:])
#         y.extend(track[1,:])
#
#         mbd.append(Miniball.ComputeMbd(x,y))
#
#     return mbd

# END FUNCTION

# FUNCTION computeems
# compute the effective mean speed of each supplied track

# DISABLED as it requires the C++ module Miniball

# def computeems(tracks, dt=1):
#     mbd = computeminbounddiameter(tracks)
#     dur = [len(x[0,:])*dt for x in tracks]
#     return [x/float(y) for (x,y) in zip(mbd, dur)]

# END FUNCTION

# FUNCTION computemeanframewisespeed

def computemeanframewisespeed(tracks, dt=1):
    tracks_rt = trackcarttopolar(tracks)
    mean_r = [np.mean(t[0]) for t in tracks]
    dur = [len(t[0,:])*dt for t in tracks]
    return [x/float(y) for (x,y) in zip(mean_r, dur)]

# END FUNCTION

# FUNCTION computemedianabscurvature
# compute the MEDIAN ABSOLUTE curvature of each supplied track
def computemedianabscurvature(tracks):
    
    c = computecurvature(tracks)
    return [np.median(np.abs(x)) for x in c]

# END FUNCTION

# FUNCTION computemeanabsoluteframewiseanglechange
def computemeanabsoluteframewiseanglechange(tracks):

    tracks_rt = trackcarttopolar(tracks)
    mean_a = [np.mean(np.abs(t[1,1:])) for t in tracks_rt]
    return mean_a

# END FUNCTION

# FUNCTION computecurvature
# compute the full list of curvature values for each supplied track
def computecurvature(tracks):
    
    # convert tracks to polar form
    tracks_rt = trackcarttopolar(tracks)
    curvatures = []
    
    for trackrt in tracks_rt:
        r = trackrt[0]
        t = trackrt[1][1:]
        sumdisp = [r[i]+r[i+1] for i in range(len(r)-1)]
        curvatures.append(np.divide(t,sumdisp))
    
    return curvatures
    
# END FUNCTION

# FUNCTION computebootstrapconfidenceintervals
# estimate the 95% two-tailed CI from bootstrap simulation of transition parameters
def computebootstrapconfidenceintervals(real_estimate, bootstrap_estimates):
	
    num_dim = len(np.shape(bootstrap_estimates))
    if num_dim==1:
        # single run
        p01 = real_estimate[0]
        p10 = real_estimate[1]
        bs01 = [x[0] for x in bootstrap_estimates]
        bs10 = [x[1] for x in bootstrap_estimates]
        return bs_conf_int(bs01, p01), bs_conf_int(bs10, p10)
    
    else:
        num_iter = len(bootstrap_estimates)
        assert len(real_estimate)==num_iter, "Number of noise levels in the real and bootstrap runs do not match"
        B = len(bootstrap_estimates[0])
        CIs01 = []
        CIs10 = []

        for i in range(num_iter):
            p01 = real_estimate[i][0]
            CIs01.append( bs_conf_int([x[0] for x in bootstrap_estimates[i]], p01) )
            
            p10 = real_estimate[i][1]
            CIs10.append( bs_conf_int([x[1] for x in bootstrap_estimates[i]], p10) )

        return CIs01, CIs10

#END FUNCTION

# FUNCTION bs_conf_int
# actual mechanism of computing bs confidence intervals
def bs_conf_int(bs, t):
    w = np.array(bs)-t
    w.sort()
    return [w[24],w[974]]

# END FUNCTION

def checktracksandrunprobs(tracks_xy, run_probs):
    
    assert len(tracks_xy)==len(run_probs), "Number of supplied tracks does not match number of supplied state probability sequences."

# END FUNCTION

# FUNCTION
def postprocessrunprobs(run_probs, run_min, stop_min):
    
    new_run_probs = []
    
    for rp in run_probs:
        
        rp_new = [1 if x>0.5 else 0 for x in rp]
        
        labels, num_runs = meas.label(rp_new)
        run_length = [sum([x==i for x in labels]) for i in range(1,num_runs+1)]
        
        for n in range(num_runs):
            if run_length[n]<run_min:
                # switch run to stop
                rp_new[np.where(labels==n+1)[0]] = 0
        
        labels, num_stops = meas.label(rp_new)
        stop_length = [sum([x==i for x in labels]) for i in range(1,num_stops+1)]

        for n in range(num_stops):
            if stop_length[n]<stop_min:
                # switch stop to run
                rp_new[np.where(labels==n+1)[0]] = 1
                
        new_run_probs.append(rp_new)
        
    return new_run_probs
# END FUNCTION

# FUNCTION heuristiccomputethreshold
def heuristiccomputethreshold(empirical_priors, lower_bound, upper_bound):
    y1 = empirical_priors[0]
    y2 = empirical_priors[2]
    d = lambda x: (y1(x)-y2(x))**2
    
    cutoff = optimize.fminbound(d, lower_bound, upper_bound)
    
    return cutoff

# END FUNCTION

# FUNCTION computeerrorrates
def computeerrorrates(numfp, numfn, numtrue):
    meantrue = [np.mean(a) for a in numtrue]
    fprate = [x/y for x,y in zip([np.mean(a) for a in numfp], meantrue)]
    fnrate = [x/y for x,y in zip([np.mean(a) for a in numfn], meantrue)]
    fpstd = [x/y for x,y in zip([np.std(a) for a in numfp], meantrue)]
    fnstd = [x/y for x,y in zip([np.std(a) for a in numfn], meantrue)]
    return fprate, fnrate, fpstd, fnstd
# END FUNCTION
