#include <Python.h>
#include "string.h"
#include <vector>
#include <numeric>
#include <cmath>
#include <iostream>
#include <map>

const double PI = acos(-1.0);
const double Tolerance = 1e-15;
const double Eps = 1e-40;

using namespace std;

// function declarations
double linear_interp1(double XIn, map<double, double> *Interp);
double hmm_main(vector<double> *R, vector<double> *Theta, double P01,
		double P10, map<double, double> *Displacement0,
		map<double, double> *Displacement1, map<double, double> *Angle0,
		map<double, double> *Angle1, vector<double> *RunProbs);

double linear_interp1(double XIn, map<double, double> *Interp) {

	/**
	 * Check first whether lookup value is in range.
	 * If not then just return first or last value.
	 */

	map<double, double>::iterator EndOfData = Interp->end();
	EndOfData--;
	double XMax = EndOfData->first;
	double XMin = Interp->begin()->first;

	if (XIn <= XMin) {
		return Interp->begin()->second;
	}
	if (XIn >= XMax) {
		return EndOfData->second;
	}

	// find the interval in which XIn lies
	map<double, double>::iterator LowFind, LowFindMinusOne;

	LowFind = Interp->lower_bound(XIn);
	LowFindMinusOne = LowFind;
	LowFindMinusOne--;

	// linear interpolation

	double x0, x1, y0, y1, YOut;

	y0 = LowFindMinusOne->second;
	y1 = LowFind->second;
	x0 = LowFindMinusOne->first;
	x1 = LowFind->first;

	YOut = y0 + (XIn - x0) / (x1 - x0) * (y1 - y0);

	return YOut;

}

// END double linear_interp1(double XIn, map<double, double> *Interp)

double hmm_main(vector<double> *R, vector<double> *Theta, double P01,
		double P10, map<double, double> *Displacement0,
		map<double, double> *Displacement1, map<double, double> *Angle0,
		map<double, double> *Angle1, vector<double> *RunProbs) {

	// check whether run probs are required
	bool bRunProbs;
	if (RunProbs) {
		bRunProbs = true; }
	else {
		bRunProbs = false; }

	int DebugNum = 0;

	unsigned int NumTransitions = R->size();
	double NegLogLikelihood = 0.0;
	double Transition[2] = { P01, P10 };
	double P00 = 1 - P01;
	double P11 = 1 - P10;

	vector<double> Alpha0, Alpha1;
	vector<double> Beta0, Beta1;
	Alpha0.resize(NumTransitions);
	Alpha1.resize(NumTransitions);
	Beta0.resize(NumTransitions);
	Beta1.resize(NumTransitions);

	// initial values
	Beta0.at(NumTransitions - 1) = 1.0;
	Beta1.at(NumTransitions - 1) = 1.0;

	double r = R->at(0);
	double t = Theta->at(0);

	Alpha0.at(0) = linear_interp1(r, Displacement0);
	Alpha1.at(0) = linear_interp1(r, Displacement1);

	for (unsigned int j = 1; j < NumTransitions; ++j) {

//		cout << DebugNum++ << endl;

// ALPHA COMPUTATION

		r = R->at(j);
		t = Theta->at(j);

		double ObsSpeedx0, ObsSpeedx1, ObsAngle11, ObsAnglex0, ObsAngle01, b00,
				b01, b10, b11;
		ObsSpeedx0 = linear_interp1(r, Displacement0);
		ObsSpeedx1 = linear_interp1(r, Displacement1);
		ObsAnglex0 = linear_interp1(t, Angle0);
		ObsAngle01 = ObsAnglex0;
		ObsAngle11 = linear_interp1(t, Angle1);

		b00 = ObsSpeedx0 * ObsAnglex0;
		b01 = ObsSpeedx1 * ObsAngle01;
		b10 = ObsSpeedx0 * ObsAnglex0;
		b11 = ObsSpeedx1 * ObsAngle11;

//		cout << "ALPHA" <<endl;
//		cout << "b00[" << j << "]=" << b00 << endl;
//		cout << "b01[" << j << "]=" << b01 << endl;
//		cout << "b10[" << j << "]=" << b10 << endl;
//		cout << "b11[" << j << "]=" << b11 << endl;

// fix very small values
//		ObsAngle01 = (ObsAngle01<Eps) ? Eps : ObsAngle01;
//		ObsAngle11 = (ObsAngle11<Eps) ? Eps : ObsAngle11;
//		ObsAnglex0 = (ObsAnglex0<Eps) ? Eps : ObsAnglex0;
//		ObsSpeedx0 = (ObsSpeedx0<Eps) ? Eps : ObsSpeedx0;
//		ObsSpeedx1 = (ObsSpeedx1<Eps) ? Eps : ObsSpeedx1;

		Alpha0.at(j) = Alpha0.at(j - 1) * P00 * b00
				+ Alpha1.at(j - 1) * P10 * b10;

		Alpha1.at(j) = Alpha0.at(j - 1) * P01 * b01
				+ Alpha1.at(j - 1) * P11 * b11;

		// renormalise every step

		double RenormConst = Alpha0.at(j) + Alpha1.at(j);
		NegLogLikelihood -= log(RenormConst);

		Alpha0.at(j) = Alpha0.at(j) / RenormConst;
		Alpha1.at(j) = Alpha1.at(j) / RenormConst;

		// replace any values below NormTolerance with zero
//		if (Alpha0.at(j)<Eps)
//			Alpha0.at(j) = 0.0;
//
//		if (Alpha1.at(j)<Eps)
//				Alpha1.at(j) = 0.0;

//		cout << "Alpha loop" << endl;

// BETA COMPUTATION

		if (bRunProbs) {

			unsigned int BetaIdx = NumTransitions - j;
			r = R->at(BetaIdx);
			t = Theta->at(BetaIdx);

			ObsSpeedx0 = linear_interp1(r, Displacement0);
			ObsSpeedx1 = linear_interp1(r, Displacement1);
			ObsAnglex0 = linear_interp1(t, Angle0);
			ObsAngle01 = ObsAnglex0;
			ObsAngle11 = linear_interp1(t, Angle1);

			b00 = ObsSpeedx0 * ObsAnglex0;
			b01 = ObsSpeedx1 * ObsAngle01;
			b10 = ObsSpeedx0 * ObsAnglex0;
			b11 = ObsSpeedx1 * ObsAngle11;

//		cout << "BETA" <<endl;
//		cout << "b00[" << j << "]=" << b00 << endl;
//		cout << "b01[" << j << "]=" << b01 << endl;
//		cout << "b10[" << j << "]=" << b10 << endl;
//		cout << "b11[" << j << "]=" << b11 << endl;

// fix very small values
//		ObsAngle01 = (ObsAngle01<Eps) ? Eps : ObsAngle01;
//		ObsAngle11 = (ObsAngle11<Eps) ? Eps : ObsAngle11;
//		ObsAnglex0 = (ObsAnglex0<Eps) ? Eps : ObsAnglex0;
//		ObsSpeedx0 = (ObsSpeedx0<Eps) ? Eps : ObsSpeedx0;
//		ObsSpeedx1 = (ObsSpeedx1<Eps) ? Eps : ObsSpeedx1;

			Beta0.at(BetaIdx - 1) = Beta0.at(BetaIdx) * P00 * b00
					+ Beta1.at(BetaIdx) * P01 * b01;

			Beta1.at(BetaIdx - 1) = Beta0.at(BetaIdx) * P10 * b10
					+ Beta1.at(BetaIdx) * P11 * b11;

			RenormConst = Beta0.at(BetaIdx - 1) + Beta1.at(BetaIdx - 1);

			Beta0.at(BetaIdx - 1) = Beta0.at(BetaIdx - 1) / RenormConst;
			Beta1.at(BetaIdx - 1) = Beta1.at(BetaIdx - 1) / RenormConst;

//		cout << "Beta loop" << endl;
		}

	} // end loop over transitions

	if (bRunProbs) {

		DebugNum = 0;

		for (unsigned int j = 0; j < Alpha0.size(); ++j) {

//		cout << DebugNum++ << endl;
//		if (j==1) {
//			cout << "Alpha0.at(" << j << ") = " << Alpha0.at(j) <<endl;
//			cout << "Alpha1.at(" << j << ") = " << Alpha1.at(j) <<endl;
//			cout << "Beta0.at(" << j << ") = " << Beta0.at(j) <<endl;
//			cout << "Beta1.at(" << j << ") = " << Beta1.at(j) <<endl;
//		}

			double a = Alpha1.at(j) * Beta1.at(j);
			double b = Alpha0.at(j) * Beta0.at(j) + Alpha1.at(j) * Beta1.at(j);

			RunProbs->at(j) = a / b;

		}

	}

	return NegLogLikelihood;

}

/** END double hmm_main(vector<double> *R, vector<double> *Theta, double P01, double P10, map<double, double> *Displacement0,
 map<double, double> *Displacement1, map<double, double> *Angle0, map<double, double> *Angle1, vector<double> *RunProbs) **/

