#!/usr/bin/env python

'''
Created on 5 Apr 2013

@author: gabs

Contains usage documentation and examples for simulation study
'''

from bayesian_track_analysis import io, process, plot, hmm
import plot.histogram as grplt
from scipy.interpolate import interp1d
from scipy.stats import weibull_min
import matplotlib.pyplot as plt
import numpy as np
import simulate
import pdb
import pickle

PI = np.pi

############################################
### STAGE 1: load data and fit params
############################################

input_directory = "/home/gabs/Dropbox/2020_internship/test_data/"
csv_filename = input_directory+"1353.csv"

# load tracks
tracks_nc = io.loadtrackscsv(csv_filename)

# compute displacements
tracks_rt = process.trackcarttopolar(tracks_nc)
displacements = np.concatenate([d[0,:] for d in tracks_rt])

# fit standard 2-parameter Weibull distribution to this
# 'location' parameter shifts the distribution and is zero in this case
shape,loc,scale=weibull_min.fit(displacements*0.12*50,floc=0) # units um s^-1

# define run speed distribution and stopwise angle change distribution
runspeedrng = lambda x : scale*np.random.weibull(shape)
anglechangerng = lambda x : 2*PI*np.random.rand()-PI

# simulate some noise free tracks WT
numtracks = 500
dt = 0.02
timetotal = 5
meanrunduration = 1.
meanstopduration = 0.1

noiselevels = np.concatenate([[0.01], np.arange(0.2,2.2,0.2)]) * (0.12**2) * 50 # units um^2 s^-1

numfp_full = []
numfn_full = []
numtrue_full = []
matchesidx_full = []
tps_full = []
numfp_so = []
numfn_so = []
numtrue_so = []
matchesidx_so = []
tps_so = []

numfp_full_pp = []
numfn_full_pp = []
numtrue_full_pp = []
matchesidx_full_pp = []
numfp_so_pp = []
numfn_so_pp = []
numtrue_so_pp = []
matchesidx_so_pp = []

numfp_heur = []
numfn_heur = []
numtrue_heur = []
matchesidx_heur = []
numfp_heur_pp = []
numfn_heur_pp = []
numtrue_heur_pp = []
matchesidx_heur_pp = []
heur_cutoff = []

for noiselevel in noiselevels:
    
    time, sim_tracks_wt, eventtime, eventposition = simulate.simulaterunandstopcrw(numtracks, dt, timetotal, meanrunduration, meanstopduration, runspeedrng, anglechangerng, noiselevel)
    
    # NC
    temp = simulate.simulaterunandstopcrw(numtracks, dt, timetotal, 1e12, 0., runspeedrng, anglechangerng, noiselevel)
    sim_tracks_nc = temp[1]
    
    # NM
    temp = simulate.simulaterunandstopcrw(numtracks, dt, timetotal, 1e-12, 1e12, runspeedrng, anglechangerng, noiselevel)
    sim_tracks_nm = temp[1]
    
    # discretise true state sequence
    truestatus = process.discretiseunderlyingstatesequence(eventtime, time)
    
    # run HMM
    empirical_priors = process.computepriors(sim_tracks_nm, sim_tracks_nc, sim_tracks_wt)
    
    # tracks_wt is a list containing the unmodified tracks loaded from CSV
    # tracks_wt_rt is the modified polar representation
    sim_tracks_wt_rt = process.trackcarttopolar(sim_tracks_wt)
    
    # heuristic
    hc = process.heuristiccomputethreshold(empirical_priors, 0, 1.)
    heur_cutoff.append(hc)
    
    run_probs = [[1 if x<hc else 0 for x in t[0]] for t in sim_tracks_wt_rt] 
    
    # test output against ground truth
    thisnumfp,thisnumfn,thisnumtrue,thismatchesidx = process.compareinferredandtruesequences(run_probs, np.copy(truestatus))
    
    numfp_heur.append(thisnumfp)
    numfn_heur.append(thisnumfn)
    numtrue_heur.append(thisnumtrue)
    matchesidx_heur.append(thismatchesidx)
    
    # post-processing
    run_status = process.postprocessrunprobs(run_probs, 2, 2)
    thisnumfp,thisnumfn,thisnumtrue,thismatchesidx = process.compareinferredandtruesequences(run_status, np.copy(truestatus))

    numfp_heur_pp.append(thisnumfp)
    numfn_heur_pp.append(thisnumfn)
    numtrue_heur_pp.append(thisnumtrue)
    matchesidx_heur_pp.append(thismatchesidx)
    
    # optimise the transition probabilities
    transition_probs = hmm.optimisetransitionparameters(empirical_priors, sim_tracks_wt_rt)
    
    # use these to segment tracks
    run_probs = hmm.Hmm(empirical_priors, sim_tracks_wt_rt, transition_probs, False)[0]
    
    # test output against ground truth
    thisnumfp,thisnumfn,thisnumtrue,thismatchesidx = process.compareinferredandtruesequences(run_probs, np.copy(truestatus))
    
    numfp_full.append(thisnumfp)
    numfn_full.append(thisnumfn)
    numtrue_full.append(thisnumtrue)
    matchesidx_full.append(thismatchesidx)
    tps_full.append(transition_probs)
    
    # post-processing
    run_status = process.postprocessrunprobs(run_probs, 2, 2)
    thisnumfp,thisnumfn,thisnumtrue,thismatchesidx = process.compareinferredandtruesequences(run_status, np.copy(truestatus))

    numfp_full_pp.append(thisnumfp)
    numfn_full_pp.append(thisnumfn)
    numtrue_full_pp.append(thisnumtrue)
    matchesidx_full_pp.append(thismatchesidx)
    
    # change to so method
    speed_only_angle_pdf = interp1d(np.pi*np.array([-1, 1]), 1/(2*np.pi)*np.ones(2), kind='zero')
    empirical_priors = empirical_priors[0], speed_only_angle_pdf, empirical_priors[2], speed_only_angle_pdf
    
     # optimise the transition probabilities
    transition_probs = hmm.optimisetransitionparameters(empirical_priors, sim_tracks_wt_rt)
    
    # use these to segment tracks
    run_probs = hmm.Hmm(empirical_priors, sim_tracks_wt_rt, transition_probs, False)[0]
    
    # test output against ground truth
    thisnumfp,thisnumfn,thisnumtrue,thismatchesidx = process.compareinferredandtruesequences(run_probs, np.copy(truestatus))
    
    numfp_so.append(thisnumfp)
    numfn_so.append(thisnumfn)
    numtrue_so.append(thisnumtrue)
    matchesidx_so.append(thismatchesidx)
    tps_so.append(transition_probs)    

    # post-processing
    run_status = process.postprocessrunprobs(run_probs, 2, 2)
    thisnumfp,thisnumfn,thisnumtrue,thismatchesidx = process.compareinferredandtruesequences(run_status, np.copy(truestatus))
    
    numfp_so_pp.append(thisnumfp)
    numfn_so_pp.append(thisnumfn)
    numtrue_so_pp.append(thisnumtrue)
    matchesidx_so_pp.append(thismatchesidx)
    
# reprocess variables to obtain mean values
fprate_full = [x/y for x,y in zip([np.mean(a) for a in numfp_full], [np.mean(a) for a in numtrue_full])]
fprate_full_pp = [x/y for x,y in zip([np.mean(a) for a in numfp_full_pp], [np.mean(a) for a in numtrue_full_pp])]
fprate_so = [x/y for x,y in zip([np.mean(a) for a in numfp_so], [np.mean(a) for a in numtrue_so])]
fprate_so_pp = [x/y for x,y in zip([np.mean(a) for a in numfp_so_pp], [np.mean(a) for a in numtrue_so_pp])]
fprate_heur = [x/y for x,y in zip([np.mean(a) for a in numfp_heur], [np.mean(a) for a in numtrue_heur])]
fprate_heur_pp = [x/y for x,y in zip([np.mean(a) for a in numfp_heur_pp], [np.mean(a) for a in numtrue_heur_pp])]

fnrate_full = [x/y for x,y in zip([np.mean(a) for a in numfn_full], [np.mean(a) for a in numtrue_full])]
fnrate_full_pp = [x/y for x,y in zip([np.mean(a) for a in numfn_full_pp], [np.mean(a) for a in numtrue_full_pp])]
fnrate_so = [x/y for x,y in zip([np.mean(a) for a in numfn_so], [np.mean(a) for a in numtrue_so])]
fnrate_so_pp = [x/y for x,y in zip([np.mean(a) for a in numfn_so_pp], [np.mean(a) for a in numtrue_so_pp])]
fnrate_heur = [x/y for x,y in zip([np.mean(a) for a in numfn_heur], [np.mean(a) for a in numtrue_heur])]
fnrate_heur_pp = [x/y for x,y in zip([np.mean(a) for a in numfn_heur_pp], [np.mean(a) for a in numtrue_heur_pp])]

fpstd_full = [x/y for x,y in zip([np.std(a) for a in numfp_full], [np.mean(a) for a in numtrue_full])]
fpstd_full_pp = [x/y for x,y in zip([np.std(a) for a in numfp_full_pp], [np.mean(a) for a in numtrue_full_pp])]
fpstd_so = [x/y for x,y in zip([np.std(a) for a in numfp_so], [np.mean(a) for a in numtrue_so])]
fpstd_so_pp = [x/y for x,y in zip([np.std(a) for a in numfp_so_pp], [np.mean(a) for a in numtrue_so_pp])]
fpstd_heur = [x/y for x,y in zip([np.std(a) for a in numfp_heur], [np.mean(a) for a in numtrue_heur])]
fpstd_heur_pp = [x/y for x,y in zip([np.std(a) for a in numfp_heur_pp], [np.mean(a) for a in numtrue_heur_pp])]

fnstd_full = [x/y for x,y in zip([np.std(a) for a in numfn_full], [np.mean(a) for a in numtrue_full])]
fnstd_full_pp = [x/y for x,y in zip([np.std(a) for a in numfn_full_pp], [np.mean(a) for a in numtrue_full_pp])]
fnstd_so = [x/y for x,y in zip([np.std(a) for a in numfn_so], [np.mean(a) for a in numtrue_so])]
fnstd_so_pp = [x/y for x,y in zip([np.std(a) for a in numfn_so_pp], [np.mean(a) for a in numtrue_so_pp])]
fnstd_heur = [x/y for x,y in zip([np.std(a) for a in numfn_heur], [np.mean(a) for a in numtrue_heur])]
fnstd_heur_pp = [x/y for x,y in zip([np.std(a) for a in numfn_heur_pp], [np.mean(a) for a in numtrue_heur_pp])]